/*
*********************************************
*											*
*		Astro TIME functions				*
*											*
*********************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "atime.h"
#include "dt_tbl.h"
#include "xfh_tcb.h"


/* const TT to TCG */
#define		CONST_LGG	(CONST_LG / (1.0 - CONST_LG))
/* const TDB to TCB */
#define		CONST_LBB	(CONST_LB / (1.0 - CONST_LB))

/*
---------------------------------------------
		leap second table adr.
---------------------------------------------
*/
LEAPSEC *pgdAT = NULL;

/*
#############################################
#											#
#			Date to Julain functions		#
#											#
#############################################
*/
/*
=============================================
=		date to julian (long jd)			=
=ent:	ADATE *pAdata = date				=
=exit:	LONGJD *pLjd in long JD			=
=============================================
*/
void date_julian_ljd(ADATE *pAdate, LONGJD *pLjd)
{
	double da, db;
	double dYear;
	int nMonth;

	dYear = pAdate->dYear;
	nMonth = pAdate->nMonth;
	if (nMonth < 3){
		nMonth += 12;
		dYear -= 1.0;
	}

	/* Calc. body */
	da = floor(dYear/100.0);
	db = 2.0 - da + floor(da / 4.0);
	pLjd->dDate = floor(365.25 * dYear) + floor(30.6001 * (nMonth + 1)) + db + 1720994.5;

	/* Separate Date & Time */
	pLjd->dTime = pAdate->dDFrac;
	pLjd->dDate += pAdate->nDay;

	/* Check Julian */
	if (pLjd->dDate < 2299160.5){
		pLjd->dDate -= db;
	}
}

/*
=============================================
=		date to julian (long jd)			=
=ent:	DATETIME *pADT = date & time	=
=exit:	LONGJD *pLjd in long JD			=
=============================================
*/
void dtime_julian_ljd(DATETIME *pADT, LONGJD *pLjd)
{
	pADT->date.dDFrac = pADT->time.nHour / DAYHOUR + pADT->time.nMin / DAYMIN + pADT->time.dSec / DAYSEC;
	date_julian_ljd(&pADT->date, pLjd);
}

/*
=============================================
=			date to julian(double)		=
=ent:	ADATE *pAdata = date				=
=exit:	double JD							=
=============================================
*/
double date_julian(ADATE *pAdate)
{
	LONGJD	lJD;

	date_julian_ljd(pAdate, &lJD);

	return (lJD.dDate + lJD.dTime);
}

/*
=============================================
=			date to julian(double)		=
=ent:	DATETIME *pADT = date & time	=
=exit:	double JD							=
=============================================
*/
double dtime_julian(DATETIME *pADT)
{
	LONGJD	lJD;

	pADT->date.dDFrac = pADT->time.nHour / DAYHOUR + pADT->time.nMin / DAYMIN + pADT->time.dSec / DAYSEC;
	date_julian_ljd(&pADT->date, &lJD);

	return (lJD.dDate + lJD.dTime);
}


/*
#############################################
#											#
#			Julian to Date functions		#
#											#
#############################################
*/
/*
=============================================
=		Julian to date (long jd)			=
=ent:	LONGJD *pLjd = long JD			=
=exit:	ADATE *pAdate = date				=
=============================================
*/
void juian_adate_ljd(LONGJD *pLjd, ADATE *pAdate)
{
	double dY;
	int nM;
	double da, db, dd;
	double dJDi;
	LONGJD nLJd = *pLjd;

	/* round 0.000 000s */
	da = nLJd.dTime * (DAYSEC * 1000000.0);
	db = floor(da + 0.5);
	nLJd.dTime = db / (DAYSEC * 1000000.0);
	ljd_normalize(&nLJd);


	/* start julian to date */
	dJDi = nLJd.dDate + 0.5;

	if (dJDi > 2299160.0){
		da = floor((dJDi - 1867216.25) / 36524.25);
		dJDi += (1.0 + da - floor(da / 4.0));
	}

	db = dJDi + 1524.0;
	dY = floor((db - 122.1) / 365.25);
	dd = db - floor(dY * 365.25);
	nM = (int)(floor(dd / 30.6001));

	/* Write Day & time */
	pAdate->nDay = (int)(dd - floor(nM * 30.6001));
	pAdate->dDFrac = nLJd.dTime;

	/* Write Month */
	if (nM < 14){
		pAdate->nMonth = nM - 1;
	}else{
		pAdate->nMonth = nM - 13;
	}

	/* Write Year */
	if (pAdate->nMonth > 2){
		pAdate->dYear = dY - 4716.0;
	}else{
		pAdate->dYear = dY - 4715.0;
	}
}

/*
=============================================
=		Julian to date (long jd)			=
=ent:	LONGJD *pLjd = long JD			=
=exit:	DATETIME *pADT = date & time	=
=============================================
*/
void juian_dtime_ljd(LONGJD *pLjd, DATETIME *pADT)
{
	juian_adate_ljd(pLjd, &pADT->date);
	dayfrac_time(pADT->date.dDFrac, &pADT->time);
}

/*
=============================================
=			Julian to date				=
=ent:	LONGJD *pLjd = long JD			=
=exit:	ADATE *pAdate = date				=
=============================================
*/
void juian_date(double dJD, ADATE *pAdate)
{
	LONGJD lJD;

	lJD.dDate = floor(dJD) + 0.5;
	lJD.dTime = dJD - lJD.dDate;
	juian_adate_ljd(&lJD, pAdate);
}

/*
=============================================
=			Julian to date				=
=ent:	LONGJD *pLjd = long JD			=
=exit:	DATETIME *pADT = date & time	=
=============================================
*/
void juian_dtime(double dJD, DATETIME *pADT)
{
	LONGJD lJD;

	lJD.dDate = floor(dJD) + 0.5;
	lJD.dTime = dJD - lJD.dDate;
	juian_dtime_ljd(&lJD, pADT);
}

/*
=============================================
=			julian to period day			=
=============================================
*/
int jd_period(double dJD, int nPerioDay)
{
	int nJD = (int)floor(dJD + 0.5);
	int nP = nJD % nPerioDay;
	if (nP < 0){
		nP += nPerioDay;
	}

	return nP;
}


/*
=============================================
=		daytime to time structer			=
=============================================
*/
void dayfrac_time(double dDayFrac, ATIME *pAtime)
{
	double dTemp;

	dTemp = dDayFrac * (DAYSEC * 1000000.0);
	dTemp = floor(dTemp + 0.5);
	dTemp /= 1000000.0;

	pAtime->nHour = (int)floor(dTemp/3600.0);
	dTemp -= pAtime->nHour * 3600.0;

	pAtime->nMin = (int)floor(dTemp/60.0);
	dTemp -= pAtime->nMin * 60.0;

	pAtime->dSec = dTemp;
}


/*
#############################################
#											#
#		UTC <-> TAI or TT functions		#
#											#
#############################################
*/
/*
=============================================
=	dTT(TT-UTC) function (from 1972)	=
=============================================
*/
/*
---------------------------------------------
-		set dTT table adr.				-
---------------------------------------------
*/
void set_dtt_table(LEAPSEC *pdTbl)
{
	pgdAT = pdTbl;
}

/*
---------------------------------------------
-		get dTT from UTC(Date only)		-
-											-
-entry: double dJD = JD_UTC(date only)	-
-exit:	return double dT(dayunit)		-
---------------------------------------------
*/
double get_dtt_1972(double dJD_utc)
{
	LEAPSEC *pRTbl;

	pRTbl = pgdAT;
	while( dJD_utc < pRTbl->dJD){
		pRTbl++;
	}
	return (pRTbl->dTime);
}

/*
---------------------------------------------
-		get dTT from TT					-
-											-
-ent: 	double dJD_tt = JD_TT			-
-exit:	return double dT(dayunit)		-
-		*pn60s = 0:No 60s 1:60s			-
---------------------------------------------
*/
double get_dtt_tt_1972(double dJD_tt, int *pn60s)
{
	LEAPSEC *ptbl;
	double dDTT; 

	/* search table */
	ptbl = pgdAT;
	while( dJD_tt < (ptbl->dJD + ptbl->dTime)){
		ptbl++;
	}
	dDTT = ptbl->dTime;

	/* check 60s data */
	if (pn60s != NULL){
		*pn60s = 0;
		if (ptbl != pgdAT){
			ptbl--;
			if ((dJD_tt - dDTT) >= ptbl->dJD){
				/* 60s unit */
				*pn60s = 1;
			}
		}
	}
	
	return dDTT;
}

/*
=============================================
=		get dT table data(1860-1960)	=
=											=
=ent: 	double dJD = JD_UTC				=
=exit:	return double dT(dayunit)		=
=============================================
*/
double get_dt_table(double dJD)
{
	LEAPSEC *ptbl;
	double dTemp1, dTemp2;
	int i;

	/* about jd to year */
	i = (int)((dJD - 2400410.5) / GREGOR_YEARDAY) - 1;
	if (i < 0){
		i = 0;
	}

	/* scan table */
	ptbl = &tTimesDT[i];
	while (dJD >= ptbl->dJD){
		ptbl++;
	}
	ptbl--;

	/* calc. interpolation */
	dTemp1 = (ptbl + 1)->dJD - ptbl->dJD;
	dTemp2 = (ptbl + 1)->dTime - ptbl->dTime;
	dTemp2 /= dTemp1;

	dTemp1 = ptbl->dTime + (dJD - ptbl->dJD) * dTemp2;

	return (dTemp1 / DAYSEC);
}


/*
=============================================
=		get DeltaT data approxmate		=
=											=
=ent:	double dJD = JD_UTC				=
=exit:	return double deltaT(dayunit)	=
=============================================
*/
double get_dt_approx(double dJD)
{
	double **pDATbl;
	double *pDTbl;
	double dT, dY;
	double dDT;

	/* scan table */
	pDATbl = pDT;
	while(*pDATbl != NULL){
		if(dJD >= **pDATbl){
			break;
		}
		pDATbl++;
	}
	if (*pDATbl == NULL){
		/* pre -500 */
		dDT = dJD;
		if (dJD < 625307.5L){
			dDT = 625307.5;
		}
		dY = (dDT - 2385800.5) / (JULIAN_YEARDAY*100);
		dDT = -20.0 + 32.0 * dY * dY;
		return (dDT / DAYSEC);
	}
	
	/* calc Y */
	pDTbl = *pDATbl;
	pDTbl++;
	dY = (double)dJD - *pDTbl++;
	if (*pDTbl != 0.0){
		dY *= *pDTbl;
	}
	pDTbl++;

	/* calc Cn */
	dDT = *pDTbl++;
	dT = dY;
	while(*pDTbl != 0.0){
		dDT += (*pDTbl++ * dT);
		dT *= dY;
	}

	return (dDT / DAYSEC);
}

/*
=============================================
=		get dT	(all year)					=
=											=
=ent: 	double dJD = JD_UTC				=
=exit:	return double dT(dayunit)		=
=============================================
*/
double get_dt(double dJD)
{
	/* cal. parameter table */
	if (dJD >= JD1972){
		/* (from1971) */
		return get_dtt_1972((double)dJD);
	}
	if (dJD >= 2400410.5){
		/* 1860-1971 */
		return get_dt_table(dJD);
	}

	/* other NASA */
	return get_dt_approx(dJD);
}

/*
=============================================
=			UTC to TAI or TT				=
=											=
=ent:	DATETIME *pUTC = UTC date&time	=
=exit:	return double TT-UTC sec. unit	=
=		pdJDtt in TAI or TT				=
=============================================
*/
double utc_tt(DATETIME *pUTC, LONGJD *pdJDtt)
{
	LONGJD ldJD;
	double ddAT;
	double dTimeDay;

	/*	make date to julian	& time_day */
	date_julian_ljd(&pUTC->date, &ldJD);
	dTimeDay = pUTC->time.nHour / 24.0 + pUTC->time.nMin / 1440.0 + pUTC->time.dSec / DAYSEC;

	/* get delta AT	*/
	if(ldJD.dDate >= JD1972){
		ddAT = get_dtt_1972(ldJD.dDate);
	}else{
		ddAT = get_dt(ldJD.dDate + dTimeDay);
	}

	/* add deltaAT */
	pdJDtt->dDate = ldJD.dDate;
	pdJDtt->dTime = dTimeDay + ddAT;
	ljd_normalize(pdJDtt);

	return (ddAT * DAYSEC);
}

/*
=============================================
=		UTC to TT (julian day version)	=
=											=
=ent:	double dJD_utc = UTC julianday	=
=		double *pdJD_tt = TAI storeadr.	=
=exit:	return double TT-UTC sec. unit	=
=		pdJDtt in TAI						=
=============================================
*/
double utc_tt_jd(LONGJD *pdJDutc, LONGJD *pdJDtt)
{
	double ddAT;

	/* get delta AT	*/
	ddAT = get_dt(pdJDutc->dDate + pdJDutc->dTime);

	/* add deltaAT */
	pdJDtt->dDate = pdJDutc->dDate;
	pdJDtt->dTime = pdJDutc->dTime + ddAT;
	ljd_normalize(pdJDtt);

	return (ddAT * DAYSEC);
}

/*
=============================================
=			TT to UTC						=
=											=
=ent:	double *pdJD_tt = TT data		=
=exit:	return double TAI-UTC sec. unit	=
=		pUTC in UTC date & time			=
=============================================
*/
double tt_utc(LONGJD *pdJD_tt, DATETIME *pUTC)
{
	LONGJD lJD;
	double dJD;
	double ddAT, ddATs;
	int n60sec = 0;

	dJD = pdJD_tt->dDate + pdJD_tt->dTime;
	if (dJD >= (JD1972 + (10 + DTAI_TT) / DAYSEC)){
		ddAT = get_dtt_tt_1972(dJD, &n60sec);
		if (n60sec != 0){
			pdJD_tt->dTime -= (1.0/DAYSEC);
		}
	}else{
		/* -1971 */
		ddAT = 0;
		do{
			ddATs = ddAT;
			ddAT = get_dt(dJD - ddATs);
		}while(fabs(ddATs - ddAT) > (0.1 / DAYSEC));
	}

	/*	calc. TT to UTC Julain day */
	lJD.dDate = pdJD_tt->dDate;
	lJD.dTime = pdJD_tt->dTime - ddAT;
	ljd_normalize(&lJD);

	/*	separate JD to date */
	juian_dtime_ljd(&lJD, pUTC);

	/* make 60s */
	pUTC->time.dSec += n60sec;

	return (ddAT * DAYSEC);
}

/*
=============================================
=		TT to UTC (julian day version)	=
=											=
=ent:	double *pdJD_tt = TT data		=
=exit:	return double TAI-UTC sec. unit	=
=		pUTC in UTC date & time			=
=============================================
*/
double tt_utc_jd(LONGJD *pdJDtt, LONGJD *pdJDutc)
{
	double dJD;
	double ddAT, ddAT2;

	if (pdJDtt->dDate >= (JD1972 + (10 + DTAI_TT) / DAYSEC)){
		ddAT = get_dtt_tt_1972(pdJDtt->dDate, NULL);
	}else{
		/* - 1971 */
		dJD = pdJDtt->dDate + pdJDtt->dTime;
		ddAT = 0;
		do{
			ddAT2 = ddAT;
			ddAT = get_dt(dJD - ddAT2);
		}while(fabs(ddAT2 - ddAT) > (0.1 / DAYSEC));
	}

	/*	calc. TT to UTC Julain day */
	pdJDutc->dDate = pdJDtt->dDate;
	pdJDutc->dTime = pdJDtt->dTime - ddAT;
	ljd_normalize(pdJDutc);

	/*	julian to date */
	return (ddAT * DAYSEC);
}

/*
#############################################
#											#
#				TCG functions				#
#											#
#############################################
*/
/*
=============================================
=			Get DeltaTCG(TCG-TT)			=
=											=
=ent:	double dJD_tt = TT julian day	=
=exit:	return double DeltaTCG			=
=============================================
*/
double tt_tcg(LONGJD *pdJDtt, LONGJD *pdJDtcg)
{
	double dFrac;
	double dInt;
	double dTemp;

	/* cal. dTCG */
	dTemp = CONST_LGG * (pdJDtt->dDate - JD1977);
	dFrac = CONST_LGG * (pdJDtt->dTime - DTAT_TT_DAY);

	dInt= (int)dTemp;
	dFrac += (dTemp - dInt);

	/* Calc. TCG */
	pdJDtcg->dDate = pdJDtt->dDate + dInt;
	pdJDtcg->dTime = pdJDtt->dTime + dFrac;
	ljd_normalize(pdJDtcg);

	return ((dFrac + dInt) * DAYSEC);
}

double tcg_tt(LONGJD *pdJDtcg, LONGJD *pdJDtt)
{
	double dFrac;
	double dInt;
	double dTemp;

	/* cal. dTCG */
	dTemp = CONST_LG * (pdJDtcg->dDate - JD1977);
	dFrac = CONST_LG * (pdJDtcg->dTime - DTAT_TT_DAY);

	dInt= (int)dTemp;
	dFrac += (dTemp - dInt);

	/* Calc. TCG */
	pdJDtt->dDate = pdJDtcg->dDate - dInt;
	pdJDtt->dTime = pdJDtcg->dTime - dFrac;
	ljd_normalize(pdJDtt);

	return ((dFrac + dInt) * DAYSEC);
}

/*
#############################################
#											#
#			TCB, TDB functions			#
#											#
#############################################
*/

/*
=============================================
=			TCB to TDB						=
=											=
=ent:	LONGJD *pdJDtcb = TCB JulianDay	=
=exit:	return double TDB-TCB			=
=		LONGJD *pdJDtdb = TDB JulianDay	=
=============================================
*/
double tcb_tdb(LONGJD *pdJDtcb, LONGJD *pdJDtdb)
{
	double dFrac;
	double dInt;
	double dTemp;

	/* cal. dTCG */
	dTemp = CONST_LB * (pdJDtcb->dDate - JD1977);
	dFrac = CONST_LB * (pdJDtcb->dTime * DAYSEC - DTAI_TT) - CONST_TDB0;
	dFrac /= DAYSEC;

	dInt = (int)dTemp;
	dFrac += (dTemp - dInt);

	/* Calc. TCG */
	pdJDtdb->dDate = pdJDtcb->dDate - dInt;
	pdJDtdb->dTime = pdJDtcb->dTime - dFrac;
	ljd_normalize(pdJDtdb);

	return ((dFrac + dInt) * DAYSEC);
}

/*
=============================================
=			TDB to TCB						=
=											=
=ent:	LONGJD *pdJDtdb = TDB JulianDay	=
=exit:	return double TDB-TCB			=
=		LONGJD *pdJDtcb = TCB JulianDay	=
=============================================
*/
double tdb_tcb(LONGJD *pdJDtdb, LONGJD *pdJDtcb)
{
	double dFrac;
	double dInt;
	double dTemp;

	/* cal. dTCG */
	dTemp = CONST_LBB * (pdJDtdb->dDate - JD1977);
	dFrac = CONST_LBB * (pdJDtdb->dTime * DAYSEC - DTAI_TT) - CONST_TDB0;
	dFrac /= DAYSEC;

	dInt = (int)dTemp;
	dFrac += (dTemp - dInt);

	/* Calc. TCG */
	pdJDtcb->dDate = pdJDtdb->dDate + dInt;
	pdJDtcb->dTime = pdJDtdb->dTime + dFrac;
	ljd_normalize(pdJDtdb);

	return ((dFrac + dInt) * DAYSEC);
}


/*
=============================================
=			TT to TDB Low precision		=
=											=
=ent:	LONGJD *pdJDtt = TT JulianDay	=
=exit:	return double TDB - TT			=
=		LONGJD *pdJDtdb = TDB JulianDay	=
=============================================
*/
double tt_tdb_lp(LONGJD *pdJDtt, LONGJD *pdJDtdb)
{
	double dg = DEG_RAD(357.53 + 0.9856003 * (pdJDtt->dDate - 2451545.0 + pdJDtt->dTime));
	double dDat = 0.001658 * sin(dg) + 0.000014 * sin(2.0 * dg);

	pdJDtdb->dDate = pdJDtt->dDate;
	pdJDtdb->dTime = pdJDtt->dTime + dDat / DAYSEC;
	ljd_normalize(pdJDtdb);

	return (dDat);
}

/*
=============================================
=			TDB to TT Low precision		=
=											=
=ent:	LONGJD *pdJDtdb = TDB JulianDay	=
=exit:	return double TDB-TT				=
=		LONGJD *pdJDtt = TT JulianDay	=
=============================================
*/
double tdb_tt_lp(LONGJD *pdJDtdb, LONGJD *pdJDtt)
{
	double dg = DEG_RAD(357.53 + 0.9856003 * (pdJDtdb->dDate - 2451545.0 + pdJDtdb->dTime));
	double dDat = 0.001658 * sin(dg) + 0.000014 * sin(2.0 * dg);

	pdJDtt->dDate = pdJDtdb->dDate;
	pdJDtt->dTime = pdJDtdb->dTime - dDat / DAYSEC;
	ljd_normalize(pdJDtdb);

	return (dDat);
}
