/*
#############################################
#											#
#		TCB functions for XHF2002		#
#											#
#############################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "astro.h"
#include "xfh_tcb.h"

// HF Data file
#define	HFDATA_PATH	"HF2002.DAT"
#define	HFDAT_NE		463
#define	HFDAT_NX		36
#define	HFDAT_TOTAL	(HFDAT_NE + HFDAT_NX + 1)

// Const
#define	HF_START_JDT	2305450.5
#define	HF_N1			218997.0

// HF data
double	dHf_a1[HFDAT_TOTAL];
double	dHf_a2[HFDAT_TOTAL];
double	dHf_f[HFDAT_TOTAL];
// HF const
double	dPI2;
double dNd;

/*
=============================================
=		Read XHF2002.DAT					=
=exit:	success retun 0					=
=============================================
 */
int read_hfdata(void)
{
	FILE *fp;
	int i;

	fp = fopen(HFDATA_PATH, "r");
	if (fp == NULL){
		return XHF_FILE_ERR;
	}

	for(i=0; i<HFDAT_TOTAL; i++){
		fscanf(fp, "%lf %lf %lf", &dHf_a1[i], &dHf_a2[i], &dHf_f[i]);
	}
	fclose(fp);

	// makeconst
	dPI2 = 8.0 * atan(1.0);
	dNd = 4.0 / HF_N1;

	return XHF_NO_ERR;
}

/*
=============================================
=		calc. TCG-TCB						=
=ent:	LONGJD *pLjd = TT Julianday		=
=exit:	retun TCG-TCB(sec)				=
=============================================
 */
int xhf(LONGJD *pLjd, double *pdTime)
{
	LONGJD lgTmp;
	double t;
	double dPI2t;
	double dXi;
	double s0, s1;
	double arg;
	int i, j;

	lgTmp.dDate = (pLjd->dDate - HF_START_JDT) - 0.5 * HF_N1;
	lgTmp.dTime = pLjd->dTime;
	if (lgTmp.dDate <= 0){
		return XHF_SPAN_ERR;
	}

	t = ljdtojd(&lgTmp);
	dPI2t = dPI2 * t;
	dXi = dNd * t;

	s0 = dHf_a1[0] + (dHf_a2[0] + dHf_f[0] * dXi) * dXi;

	j = HFDAT_NE + HFDAT_NX;
	s1 = 0.0;
	for(i=0; i<HFDAT_NX; i++){
		arg = dPI2t * dHf_f[j];
		s1 += (dXi * (dHf_a1[j] * sin(arg) + dHf_a2[j] * cos(arg)));
		j--;
	}

	for (i=0; i<HFDAT_NE; i++){
		arg = dPI2t * dHf_f[j];
		s1 += (dHf_a1[j] * sin(arg) + dHf_a2[j] * cos(arg));
		j--;
	}

	*pdTime = s1 + s0;

	return XHF_NO_ERR;
}

/*
=============================================
=			TT,TCG to TCB					=
=											=
=ent:	LONGJD *pdJDtt = TT JulianDay	=
=		LONGJD *pdJDtcg = TCG JulianDay	=
=exit:	return double TCB-TCG			=
=		LONGJD *pdJDtcb = TCB JulianDay	=
=============================================
*/
double xhf_tcg_tcb(LONGJD *pdJDtt, LONGJD *pdJDtcg, LONGJD *pdJDtcb)
{
	double	dTCB;

	// Get TCB - TCG
	if (xhf(pdJDtt, &dTCB) != XHF_NO_ERR){
		// time span error
		return -1;
	}

	// Calc. TCB
	pdJDtcb->dDate = pdJDtcg->dDate;
	pdJDtcb->dTime = pdJDtcg->dTime + (dTCB / DAYSEC);
	ljd_normalize(pdJDtcb);

	return dTCB;
}

/*
=============================================
=			TCB to TT,TCG					=
=											=
=ent:	LONGJD *pdJDtcb = TCB JulianDay	=
=exit:	return double TCB-TCG			=
=		LONGJD *pdJDtt = TT JulianDay	=
=		LONGJD *pdJDtcg = TCG JulianDay	=
=============================================
*/
double xhf_tcb_tt(LONGJD *pdJDtcb, LONGJD *pdJDtt, LONGJD *pdJDtcg)
{
	double	dTCB;
	double nOldDay;

	*pdJDtt = *pdJDtcb;
	do{
		nOldDay = pdJDtt->dTime;
		xhf(pdJDtt, &dTCB);
		pdJDtcg->dDate = pdJDtcb->dDate;
		pdJDtcg->dTime = pdJDtcb->dTime - (dTCB / DAYSEC);
		tcg_tt(pdJDtcg, pdJDtt);
	}while(fabs(nOldDay - pdJDtt->dTime) > 1e-15);

	return dTCB;
}


